/*

	Reader.cpp v0.8.1
    
	Copyright 1997 Jeremy Rand

This file contains the code for reading data from the modem and
turning the stream of bytes into tokens.  Also, the main reader
thread loop is in here also.

*/

#include "Reader.h"


int ReaderRunning;


void ReaderDie(long result)
{
	// This function is called when the reader thread terminates.
	// It tells any running preferences window that the reader is stopped
	
	ReaderRunning=FALSE;
	
	be_app->PostMessage(MSG_READERDEAD);
	
	exit_thread(result);
	}
		

CNDReader::CNDReader(const char *portname, const char *initstring, 
			data_rate CND_data_rate, data_bits CND_data_bits, 
			parity_mode CND_parity_mode, stop_bits CND_stop_bits)
{
	// Open the serial port and set up the tokens
	
	calltime=callname=callnum=NULL;

	modem = new BSerialPort;

	if (modem->Open(portname) == B_ERROR) {
		// if we can't open the serial port, I just put up an alert and
		// quit the thread.  I could throw an exception, but I know nothing
		// needs to be freed, so I just exit.  This might change in the
		// future.
		
		BAlert *anAlert = new BAlert("",
			"An error has occured opening the serial port",
			"OK");
		anAlert->Go();
		delete modem;
		ReaderDie(1);
		}

	modem->SetTimeout(25500000.0);	// This one I would rather not have.
									// I would prefer to block forever if necessary
									// See CNDReader::GetMore for more info on this
	modem->SetBlocking(FALSE);
	modem->SetDataRate(CND_data_rate);
	modem->SetDataBits(CND_data_bits);
	modem->SetParityMode(CND_parity_mode);
	modem->SetStopBits(CND_stop_bits);

	if (modem->Write(initstring, strlen(initstring))!=strlen(initstring)) {
		// if writing the initstring to the modem fails, I put up an alert
		// just like if I can't open the serial port.  Again, I might change
		// the below to an exception some day.
		
		BAlert *anAlert = new BAlert("",
			"An error has occured sending the init string",
			"OK");
		anAlert->Go();
		modem->Close();
		delete modem;
		ReaderDie(1);
		}
		
	buffer[0]=0;					// Nothing in the buffer to start
	cursor=end=buffer;
	
	CND_Strings[CND_EQUALS]="=";	// Initialize the tokens.
	CND_Strings[CND_RING]="RING";
	
	CND_Strings[CND_ZTIME]="TIME:";
	CND_Strings[CND_ZNUMBER]="CALLER NUMBER:";
	CND_Strings[CND_ZNAME]="CALLER NAME:";
	CND_Strings[CND_ZNONUMBER]="REASON FOR NO NUMBER:";
	CND_Strings[CND_ZNONAME]="REASON FOR NO NAME:";
	
	CND_Strings[CND_OTIME]="TIME";
	CND_Strings[CND_ODATE]="DATE";
	CND_Strings[CND_ONUMBER]="NMBR";
	CND_Strings[CND_ONAME]="NAME";
	}
	
CNDReader::~CNDReader()
{
	// This function just cleans up
	
	if (calltime != NULL)
		delete[] calltime;
	if (callname != NULL)
		delete[] callname;
	if (callnum != NULL)
		delete[] callname;
	modem->Close;
	delete modem;
	}
		
	
void CNDReader::SkipWS(void)
{
	// This member function just points the cursor to the next non
	// whitespace character in the buffer
	
	while ((isspace(*cursor)) || (cursor==end)) {
		if (cursor==end) 
			GetMore();		// There are only whitespace characters left in the buffer
		else
			cursor++;		// Skip this whitespace character
		}
	}
	
void CNDReader::NextWord(void)
{
	// This member function is like SkipWS, except is skips to the next
	// word where words are separated by whitespace.
	
	while (!isspace(*cursor)) {
		if (cursor==end) 
			GetMore();		// There are only non-whitespace characters left in the buffer
		else
			cursor++;		// Skip this non-whitespace character
		}
		
	// At this point, the cursor definitely points to a whitespace character
	// To get to the next word, why duplicate code
	
	SkipWS();	
	}
	
void CNDReader::GetMore(void)
{
	// This function is where the reader thread spends most of its time.  It
	// will block in this function until one of the following happens:
	//   1. More data is received on the serial port and placed into the buffer
	//   2. ReaderRunning gets set to FALSE
	//   3. Some kind of error occurs.
	
	long bytesfree, bytesread;

	if (cursor!=buffer) {
		// This bit of code checks if the cursor is at the start of the buffer
		// If it is not, it throws out all the data in the buffer before the
		// cursor, just to avoid overflow.
		
		char *dst=buffer;
		char *src;
		
		for (src=cursor; src < end; src++, dst++)
			*dst=*src;
		cursor=buffer;
		end=dst;
		*end=0;
		}
	
	bytesfree=BUFFERSIZE - (end - buffer) - 1;
	if (bytesfree == 0) {
		// If there is no space left in the buffer, this is a serious error
		// and should not occur.  Exit cleanly.
		
		BAlert *anAlert = new BAlert("",
			"Buffer overflow has occured\n"
			"Please report the problem to the author",
			"OK");
		anAlert->Go();
		
		throw CND_Error();
		}
	
	while(1) {
		if (!ReaderRunning)
    		throw Normal_Quit();
    	
    	switch(modem->WaitForInput()) {
    		// The read above blocks for half a second (set in the CNDReader
    		// constructor.  This is so I can go back and check for a quit
    		
    		case 0:
				break;			// Nothing, check to see if we should quit.

			case B_ERROR:
				// Probably got a signal, check the ReaderRunning flag!
				break;

			default:
				// The read got some data, set the null terminator, end pointer
				// and return
				
				if ((bytesread=modem->Read(end, bytesfree)) != 0) {
					end[bytesread]=0;
					end=&end[bytesread];
					return;
					}
			}			
    	}
	}

int CNDReader::GetNextToken(void)
{
	// This function looks for each token one at a time in the buffer.  The
	// order it checks the tokens is important.  It needs to check the shortest
	// tokens first and the longest tokens last.  The CNDReader constructor creates
	// them in this order for us.
	
	int i;
	int currsize=end-cursor;
	int CND_length;
	
	if (currsize==0) {
		// If the buffer is currently empty, time to get more data
		GetMore();
		currsize=end-cursor;	// I kept forgetting to reset the currsize after doing
								// a GetMore.  More bugs squashed.
		}
	
	SkipWS();					// Make sure we are at the start of a new token
	currsize=end-cursor;		// Again, reset currsize.
								// I wonder why I don't just do a SkipWS and then
								// set currsize.  Can't remember if there is a good
								// reason to do it this way.  Oh well.
	
	while (1) {
		for(i=CND_FIRST_TOKEN; i < NUM_TOKENS; i++) {
			// Check each token one by one (smallest to largest)
			
			CND_length=strlen(CND_Strings[i]);
			if (currsize < CND_length) {
				// If there is less data in the buffer than the size of the token
				// just compare what we have in the buffer

				if (strncmp(cursor, CND_Strings[i], currsize)==0) {
					// If what we have matches, better get some more and see if we
					// have a real match here.  To do so, fiddle with the counter
					// to recheck this token (I know, bad form)

					GetMore();
					currsize=end-cursor;
					i--;
					}
				}
			else {
				// Else, we have more data in the buffer than the length of the token
				// Check for a match with the whole token
				
				if (strncmp(cursor, CND_Strings[i], CND_length)==0) {
					// We have a match on this token.  Move the cursor ahead and return
					// the token as a match.
					
					cursor+=CND_length;
					// SkipWS();  Not a good idea after all, can block returning this token
					// if a non WS char is not yet in the buffer.  This can unneccessarily
					// delay processing this token.
					return(i);
					}
				}
			}
		// If we get to here, the buffer matches no tokens.  In this case we just
		// skip to the next word and ignore it.  It will work out in the end plus
		// we would have to put in tokens for things like OK and other phrases
		// uttered by modems.
		
		NextWord();
		currsize=end-cursor;
		}
	}

	
char *CNDReader::GetRestLine(void)
{
	// This bit of code just grabs everything from the cursor until an end of line
	// character is received.
	
	int linelength;

	SkipWS();
	for(linelength=0;((cursor[linelength]!='\r')&&(cursor[linelength]!='\n'));linelength++) {
		// While we are not at an eol character yet, check next

		if (cursor[linelength]==0) {
			// But we have run out of data in the buffer, better get some more and
			// recheck this character because it is now non-null.
	
			GetMore();
			linelength--;		// I know, not very good form twiddling with the index
			}
		}
	// We got an eol character, grab the info from the buffer and return it
	
	char *line=new char[linelength+1];
	strncpy(line, cursor, linelength);
	line[linelength]=0;
	
	cursor=&cursor[linelength];		// Reset the cursor to the new position

	// SkipWS(); Not a good idea again since it can delay having the window displayed
	return(line);
	}
	

void CNDReader::GetNextCall(void)
{
	// This function will block until a call comes in.  It will set the protected
	// variables with the info from the call.  Other functions can be used to get
	// the results.
	//
	// It supports ZyXEL, Rockwell and USRobotics CallerID formats
	
	int currtok;
	char *aString, *aString2;
	
	while (1) {
	
		do {			// Do nothing until we get a time token
			currtok=GetNextToken();
			}
		while((currtok!=CND_ZTIME) && (currtok!=CND_ODATE));
		
		if (currtok==CND_ZTIME) {		// It's a ZyXEL modem
			aString=GetRestLine();
					// Store the time data
			calltime=new char[strlen(aString) + 16];
			strcpy(calltime, "Time of call: ");
			strcat(calltime, aString);
			delete[] aString;
			
			currtok=GetNextToken();
			switch (currtok) {
				case CND_ZNAME:
					aString=GetRestLine();
					callname=new char[strlen(aString) + 17];
					strcpy(callname, "Caller's name: ");
					break;
				case CND_ZNONAME:
					aString=GetRestLine();
					callname=new char[strlen(aString) + 22];
					strcpy(callname, "Reason for no name: ");
					break;
				default:
					// We didn't get a name token like we expected, free what
					// we have and try for the next call.
					
					delete[] calltime;
					continue;
				}
			strcat(callname, aString);
			delete[] aString;
				
			currtok=GetNextToken();
			switch (currtok) {
				case CND_ZNUMBER:
					aString=GetRestLine();
					callnum=new char[strlen(aString) + 19];
					strcpy(callnum, "Caller's number: ");
					break;
				case CND_ZNONUMBER:
					aString=GetRestLine();
					callnum=new char[strlen(aString) + 24];
					strcpy(callnum, "Reason for no number: ");
					break;
				default:
					// We didn't get a number token like we expected, free what
					// we have and try for the next call.
					
					delete[] calltime;
					delete[] callname;
					continue;
				}
			strcat(callnum, aString);
			delete[] aString;
			break;		// If we get here, we have gotten all the data we need.
	
			}
		else {								// Not a ZyXEL modem
			if (GetNextToken() != CND_EQUALS)
				continue;					// Should be an Equals after DATE
			aString=GetRestLine();
			if ((GetNextToken() != CND_OTIME) || (GetNextToken() != CND_EQUALS)) {
				delete[] aString;
				continue;					// Didn't get the time or the equal sign
				}
			aString2=GetRestLine();
			calltime=new char[strlen(aString) + strlen(aString2) + 19];
			strcpy(calltime, "Time of call: ");			// The following formats the time
			strncat(calltime, aString, 2);				// and the date
			strcat(calltime, "-");
			strcat(calltime, &aString[2]);
			strcat(calltime, " ");
			strncat(calltime, aString2, 2);
			strcat(calltime, ":");
			strcat(calltime, &aString2[2]);
			delete[] aString;
			delete[] aString2;
			
			if ((GetNextToken() != CND_ONUMBER) || (GetNextToken() != CND_EQUALS)) {
				delete[] calltime;
				continue;					// Didn't get the number or the equal sign
				}
			aString=GetRestLine();
			callnum=new char[strlen(aString) + 19];
			strcpy(callnum, "Caller's number: ");
			strcat(callnum, aString);		// Store the number
			delete[] aString;
			
			currtok=GetNextToken();
			if (currtok == CND_ONAME) {
				if (GetNextToken() != CND_EQUALS) {
					delete[] calltime;
					delete[] callnum;
					continue;					// Didn't get the equal sign
					}
				else {			// We got a NAME and an equal sign
					aString=GetRestLine();
					callname=new char[strlen(aString) + 17];
					strcpy(callname, "Caller's name: ");
					strcat(callname, aString);		// Store the name
					delete[] aString;
					break;			// If we get here, we have gotten all the data we need
					}
				}
			else {		// If no NAME token, assume there is no name service on the line
				callname=new char[23];
				strcpy(callname, "Caller's name: NO NAME");
				break;				// If we get here, we have gotten all the data we need
				}
			}
		}
	}
	

char *CNDReader::GetDateTime(void)
{
	char *tmp;
	
	tmp=calltime;
	calltime=NULL;
	return(tmp);
	}


char *CNDReader::GetCallerName(void)
{
	char *tmp;
	
	tmp=callname;
	callname=NULL;
	return(tmp);
	}


char *CNDReader::GetCallerNum(void)
{
	char *tmp;
	
	tmp=callnum;
	callnum=NULL;
	return(tmp);
	}


void handler(int sig)
{
	// I don't do anything with the signal, its just a way to unblock
	}
	
	
long reader_thread(void *data)
{
	// This function is the entry point for the reader thread.  It basically
	// tries to create a CNDView and when successful, puts it in a new window.
	
	CNDReader *myreader;
	BMessage *aMessage;
	team_id aTeam;
	status_t result;
	BMessenger *aMessenger;

	signal(SIGUSR1, handler);		// Set up the signal handler
		
	SerialSet *currSettings = (SerialSet *) data;
			// The serial port settings are passed to us, for more info see CallerIDd.cpp
	
	myreader=new CNDReader(currSettings->serialport, currSettings->initstring, 
		currSettings->theDataRate, currSettings->theDataBits, currSettings->theParityMode,
		currSettings->theStopBits);
	
	be_app->PostMessage(MSG_READERRUNNING);
	
	while (1) {
	
		try {
			// Currently all exceptions are thrown in GetMore, check that for more info
			myreader->GetNextCall();
			}
		catch (CNDReader::Normal_Quit) {
			delete myreader;
			ReaderDie(0);
			}
		catch (CNDReader::CND_Error) {
			delete myreader;
			ReaderDie(1);
			}
		
		aMessage = new BMessage(MSG_NEW_CALL);		// Build message to send to CallDisplay
		aMessage->AddString("Time", myreader->GetDateTime());
		aMessage->AddString("Name", myreader->GetCallerName());
		aMessage->AddString("Number", myreader->GetCallerNum());
		
		if ((aTeam=be_roster->TeamFor(CallDisplayType)) == B_ERROR) {	// If its not running
			if (be_roster->Launch(CallDisplayType, aMessage) != B_NO_ERROR) {	// Start it
				BAlert *anAlert = new BAlert("",
					"Cannot find CallDisplay executable!",
					"OK");
				anAlert->Go();						// Display Alert if we can't start it
				}
			continue;								// Wait for next call
			}
		
		aMessenger = new BMessenger(CallDisplayType, aTeam, &result);	// Build a Messenger to CallDisplay
		if (result != B_NO_ERROR) {
			BAlert *anAlert = new BAlert("",
				"Cannot connect with running CallDisplay!", "OK");
			anAlert->Go();							// Failed to make Messenger
			}
		else {
			if (aMessenger->SendMessage(aMessage) != B_NO_ERROR) {		// Send the message to CallDisplay
				BAlert *anAlert = new BAlert("",
					"Cannot send message to CallDisplay!", "OK");
				anAlert->Go();						// Failed to send the message
				}
			}
		delete aMessenger;
		delete aMessage;
		}	
	}